@@ -18,7 +18,7 @@
     License along with this library; if not, write to the Free
     Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
-    $Id: util.c,v 1.130 2003/06/12 12:59:36 fherrera Exp $
+    $Id: util.c,v 1.137 2006/07/24 10:19:58 kmaraas Exp $
 
 ***************************************************************************/
 #include <assert.h>
@@ -31,9 +31,14 @@
 #ifdef HAVE_UNISTD_H
 #  include <unistd.h>
 #endif
+
 #include "rename.h"
 #include "util.h"
 
+#ifdef G_OS_WIN32
+#include <fcntl.h>
+#endif
+
 IDL_EXPORT const char *IDL_tree_type_names[] = {
 	"IDLN_NONE",
 	"IDLN_ANY",
@@ -218,18 +223,22 @@ int IDL_parse_filename (const char *filename, const char *cpp_args,
 	extern FILE *__IDL_in;
 	FILE *input;
 	char *cmd;
-	size_t cmd_len;
 #ifdef HAVE_CPP_PIPE_STDIN
 	char *fmt = CPP_PROGRAM " " CPP_NOSTDINC " - %s%s %s < \"%s\" %s";
-	char *wd = "", *dirend;
+	char *wd;
 #else
 	char *fmt = CPP_PROGRAM " " CPP_NOSTDINC " -I- -I%s %s \"%s\" %s";
 	char *s, *tmpfilename;
 	char cwd[2048];
 	gchar *linkto;
 #endif
+
+#ifndef G_OS_WIN32
 	char *cpperrs = (parse_flags & IDLF_SHOW_CPP_ERRORS) 
 	  ? "" : "2>/dev/null";
+#else
+	char *cpperrs = "";
+#endif
 	GSList *slist;
 	int rv;
 
@@ -240,9 +249,7 @@ int IDL_parse_filename (const char *filename, const char *cpp_args,
 	}
 #endif
 
-	if (!filename ||
-	    !tree ||
-	    (tree == NULL && ns != NULL)) {
+	if (!filename || !tree) {
 		errno = EINVAL;
 		return -1;
 	}
@@ -253,28 +260,12 @@ int IDL_parse_filename (const char *filename, const char *cpp_args,
 #endif
 
 #ifdef HAVE_CPP_PIPE_STDIN
-	if ((dirend = strrchr (filename, '/'))) {
-		int len = dirend - filename + 1;
-		wd = g_malloc (len);
-		strncpy (wd, filename, len - 1);
-		wd[len - 1] = 0;
-	}
+	wd = g_path_get_dirname (filename);
 
-	cmd_len = (strlen (filename) + (*wd ? 2 : 0) + strlen (wd) +
-		  (cpp_args ? strlen (cpp_args) : 0) +
-		  strlen(cpperrs) +
-		  strlen (fmt) - 8 + 1);
-	cmd = g_malloc (cmd_len);
-	if (!cmd) {
-		errno = ENOMEM;
-		return -1;
-	}
+	cmd = g_strdup_printf (fmt, "-I", wd, cpp_args ? cpp_args : "",
+			       filename, cpperrs);
 
-	g_snprintf (cmd, cmd_len, fmt, *wd ? "-I" : "", wd,
-		    cpp_args ? cpp_args : "", filename, cpperrs);
-
-	if (dirend)
-		g_free (wd);
+	g_free (wd);
 #else
 	s = tmpnam (NULL);
 	if (s == NULL)
@@ -283,58 +274,50 @@ int IDL_parse_filename (const char *filename, const char *cpp_args,
 	if (!getcwd (cwd, sizeof (cwd)))
 		return -1;
 
-	if (*filename == '/') {
+	if (g_path_is_absolute (filename)) {
 		linkto = g_strdup (filename);
 	} else {
-		linkto = g_malloc (strlen (cwd) + strlen (filename) + 2);
-		if (!linkto) {
-			errno = ENOMEM;
-			return -1;
-		}
-		strcpy (linkto, cwd);
-		strcat (linkto, "/");
-		strcat (linkto, filename);
+		linkto = g_strconcat (cwd, "/", filename, NULL);
 	}
 
-	tmpfilename = g_malloc (strlen (s) + 3);
-	if (!tmpfilename) {
-		g_free (linkto);
-		errno = ENOMEM;
-		return -1;
-	}
-	strcpy (tmpfilename, s);
-	strcat (tmpfilename, ".c");
+	tmpfilename = g_strconcat (s, ".c", NULL);
 #ifdef HAVE_SYMLINK
 	if (symlink (linkto, tmpfilename) < 0) {
 		g_free (linkto);
 		g_free (tmpfilename);
 		return -1;
 	}
+#else
+#error Must have symlink
 #endif
 	g_free (linkto);
 
-	cmd_len = (strlen (tmpfilename) + strlen (cwd) +
-		   (cpp_args ? strlen (cpp_args) : 0) +
-		   strlen(cpperrs) +
-		   strlen (fmt) - 6 + 1);
-	cmd = g_malloc (cmd_len);
-	if (!cmd) {
-		g_free (tmpfilename);
-		errno = ENOMEM;
-		return -1;
-	}
-
-	g_snprintf (cmd, cmd_len, fmt,
-		    cwd, cpp_args ? cpp_args : "", tmpfilename, cpperrs);
+	cmd = g_strdup_printf (fmt, cwd, cpp_args ? cpp_args : "",
+			       tmpfilename, cpperrs);
 #endif
 
        /* Many versions of cpp do evil translating internal
         * strings, producing bogus output, so clobber LC_ALL */
-       putenv ("LC_ALL=C");
+	putenv ("LC_ALL=C");
 
 #ifdef HAVE_POPEN
-	input = popen (cmd, "r");
+#ifdef G_OS_WIN32
+	if (!(parse_flags & IDLF_SHOW_CPP_ERRORS)) {
+		int save_stderr = dup (2);
+		int null = open ("NUL:", O_WRONLY);
+		dup2 (null, 2);
+		input = popen (cmd, "r");
+		close (2);
+		close (null);
+		dup2 (save_stderr, 2);
+		close (save_stderr);
+	} else
+		input = popen (cmd, "r");
 #else
+	input = popen (cmd, "r");
+#endif
+#else
+	/* Surely this doesn't make any sense? */	
 	input = fopen (cmd, "r");
 #endif
 	g_free (cmd);
@@ -391,8 +374,7 @@ int IDL_parse_filename (const char *filename, const char *cpp_args,
 	__IDL_msgcb = NULL;
 
 	if (rv != 0 || !__IDL_is_okay) {
-		if (tree)
-			*tree = NULL;
+		*tree = NULL;
 
 		if (ns)
 			*ns = NULL;
@@ -403,10 +385,7 @@ int IDL_parse_filename (const char *filename, const char *cpp_args,
 	if (__IDL_flags & IDLF_PREFIX_FILENAME)
 		IDL_ns_prefix (__IDL_root_ns, filename);
 
-	if (tree)
-		*tree = __IDL_root;
-	else
-		IDL_tree_free (__IDL_root);
+	*tree = __IDL_root;
 
 	if (ns)
 		*ns = __IDL_root_ns;
@@ -431,13 +410,11 @@ int IDL_parse_filename_with_input (const char *filename,
 	GSList *slist;
 	int rv;
 
-	if (!filename || !input_cb || !tree ||
-	    (tree == NULL && ns != NULL)) {
+	if (!filename || !input_cb || !tree) {
 		errno = EINVAL;
 		return -1;
 	}
 
-
 	IDL_parse_setup(parse_flags, max_msg_level);
 
 	__IDL_msgcb = msg_cb;
@@ -481,8 +458,7 @@ int IDL_parse_filename_with_input (const char *filename,
 	__IDL_msgcb = NULL;
 
 	if (rv != 0 || !__IDL_is_okay) {
-		if (tree)
-			*tree = NULL;
+		*tree = NULL;
 
 		if (ns)
 			*ns = NULL;
@@ -498,10 +474,7 @@ int IDL_parse_filename_with_input (const char *filename,
 	if (__IDL_flags & IDLF_PREFIX_FILENAME)
 		IDL_ns_prefix (__IDL_root_ns, filename);
 
-	if (tree)
-		*tree = __IDL_root;
-	else
-		IDL_tree_free (__IDL_root);
+	*tree = __IDL_root;
 
 	if (ns)
 		*ns = __IDL_root_ns;
@@ -514,13 +487,13 @@ int IDL_parse_filename_with_input (const char *filename,
 void yyerrorl (const char *s, int ofs)
 {
 	int line = __IDL_cur_line - 1 + ofs;
-	gchar *filename = NULL;
+	gchar *freeme = NULL, *filename = NULL;
 
 	if (__IDL_cur_filename) {
 #ifdef HAVE_CPP_PIPE_STDIN
 		filename = __IDL_cur_filename;
 #else
-		filename = g_basename (__IDL_cur_filename);
+		freeme = filename = g_path_get_basename (__IDL_cur_filename);
 #endif
 	} else
 		line = -1;
@@ -530,6 +503,7 @@ void yyerrorl (const char *s, int ofs)
 
 	/* Errors are counted, even if not printed */
 	if (__IDL_max_msg_level < IDL_ERROR)
+		g_free (freeme);
 		return;
 
 	if (__IDL_msgcb)
@@ -540,12 +514,13 @@ void yyerrorl (const char *s, int ofs)
 		else
 			fprintf (stderr, "Error: %s\n", s);
 	}
+	g_free (freeme);
 }
 
 void yywarningl (int level, const char *s, int ofs)
 {
 	int line = __IDL_cur_line - 1 + ofs;
-	gchar *filename = NULL;
+	gchar *freeme = NULL, *filename = NULL;
 
 	/* Unprinted warnings are not counted */
 	if (__IDL_max_msg_level < level)
@@ -555,7 +530,7 @@ void yywarningl (int level, const char *s, int ofs)
 #ifdef HAVE_CPP_PIPE_STDIN
 		filename = __IDL_cur_filename;
 #else
-		filename = g_basename (__IDL_cur_filename);
+		freeme = filename = g_path_get_basename (__IDL_cur_filename);
 #endif
 	} else
 		line = -1;
@@ -570,6 +545,7 @@ void yywarningl (int level, const char *s, int ofs)
 		else
 			fprintf (stderr, "Warning: %s\n", s);
 	}
+	g_free (freeme);
 }
 
 void yyerror (const char *s)
@@ -1891,10 +1867,18 @@ static void IDL_tree_free_real (IDL_tree p)
 		g_free (IDL_STRING (p).value);
 		break;
 
+	case IDLN_WIDE_STRING:
+		g_free (IDL_WIDE_STRING (p).value);
+		break;
+
 	case IDLN_CHAR:
 		g_free (IDL_CHAR (p).value);
 		break;
 
+	case IDLN_WIDE_CHAR:
+		g_free (IDL_WIDE_CHAR (p).value);
+		break;
+
 	case IDLN_IDENT:
 		g_free (IDL_IDENT (p).str);
 		g_free (IDL_IDENT_REPO_ID (p));
@@ -1959,9 +1943,12 @@ void IDL_tree_free (IDL_tree p)
 	case IDLN_TYPE_TYPECODE:
 	case IDLN_FIXED:
 	case IDLN_STRING:
+	case IDLN_WIDE_STRING:
 	case IDLN_CHAR:
+	case IDLN_WIDE_CHAR:
 	case IDLN_IDENT:
 	case IDLN_CODEFRAG:
+	case IDLN_SRCFILE:
 		__IDL_tree_free (p);
 		break;
 
@@ -3365,7 +3352,7 @@ struct IDL_emit_node {
 	IDL_tree_func post;
 };
 
-static const struct IDL_emit_node * const IDL_get_IDL_emission_table (void)
+static struct IDL_emit_node * IDL_get_IDL_emission_table (void)
 {
 	static gboolean initialized = FALSE;
 	static struct IDL_emit_node table[IDLN_LAST];
@@ -3668,7 +3655,7 @@ IDL_tree_is_recursive (IDL_tree tree, gpointer dummy)
 	info.ident_list = NULL;
 	info.recursive  = FALSE;
 
-	IDL_tree_walk2 (tree, 0, IDL_WalkF_TypespecOnly, 
+	IDL_tree_walk2 (tree, NULL, IDL_WalkF_TypespecOnly, 
 			IDL_tree_is_recursive_walker_pre, 
 			IDL_tree_is_recursive_walker_post, 
 			&info);
